//
// o p t d o w n . c
//
// Multiboot program downloader for optimize's bootcable
//
// Target OS: Linux-x86-pc
// Build: gcc -Wall -o optdown optdown.c
//
// Feb. 13, 2003 by W.Nishida (http://www.skyfree.org)
//

#include <stdio.h>	// fprintf(), stderr, fopen(), fread(), fclose()
#include <sys/io.h>	// ioperm(), inb(), outb()
#include <unistd.h>	// sleep()

// Program-specific macro

#define MAX_IDENT_RETRY	6		// Retry number for identification
#define MAX_HS_RETRY	1000		// Retry number for hand-shaking

// Parallel ports definition

int parallel_base = 0x378;		// Base address of LPT port (LPT1)

#define DTR		(parallel_base)	// Data register port
#define STR		(DTR + 1)	// Status register port
#define CTR		(DTR + 2)	// Control register port

// Status register bit (STR)

#define STR_BUSY	128		// Busy

// Control register bits (CTR)

#define CTR_INTE	16		// Interrupt enable
#define CTR_SLIN	8		// Select line
#define CTR_INIT	4		// Initialize (reset)
#define CTR_AFD		2		// Auto feed
#define CTR_STB		1		// Strobe

#define IOWAIT	0x80			// I/O wait port

// Handshaking codes

#define OPT_HS_1ST	5		// 1st handshaking code
#define OPT_HS_2ND	10		// 2nd handshaking code
#define OPT_HS_3RD	5		// 3rd handshaking code
#define OPT_HS_4TH	0x5555AAAA	// 4th handshaking code
#define OPT_HS_5TH	0xFFFFFFFF	// 5th handshaking code

// Function prototypes

int opt_ident(void);
static int recv_nibble(void);
int recv_long(void);
void send_long(unsigned int val);

// File buffer

char buf[ 256 * 1024 ];			// Transfer buffer (Max 256KB)

// Program starts...

int main(int argc, char** argv) {
  int ret, len, i;
  FILE* fp;
  unsigned int* wptr = (unsigned int*) buf;

  // Check argument

  if (argc != 2) {
    fprintf(stderr, "Usage: optdown mb-file-name\n");
    return 1;
   }

  // Get I/O permissions against parallel and hardware wait ports

  ret  = ioperm(DTR, 3, 1);
  ret |= ioperm(IOWAIT, 1, 1);
  if (ret) {
    fprintf(stderr, "I/O permission failure!\n");
    fprintf(stderr, "Do you have a root privilege?\n");
    return 1;
   }

  // Identify bootcable

  fprintf(stdout, "Searching an optimize's bootcable.");
  fflush(stdout);
  i = 0;				// Initialize counter
  while (1) {
    ret = opt_ident();			// Search the device
    if (ret == 0)			// Found!
      break;
    sleep(1);				// Wait for 1 second
    fprintf(stdout, ".");
    fflush(stdout);
    i++;				// Increment counter
    if (i >= MAX_IDENT_RETRY) {		// Too many errors
      fprintf(stderr, "\nNot found!\n");
      return 1;
     }
   }
  fprintf(stdout, " [ Found ]\n");

  // Wait for acknowledge code from bootcable

  fprintf(stdout, "Waiting for an acknowledge code from bootcable.");
  fflush(stdout);
  i = 0;				// Initialize counter
  while (1) {
    if (recv_long() == OPT_HS_4TH)	// Confirmed!
      break;
    sleep(1);				// Wait for 1 second
    fprintf(stdout, ".");
    fflush(stdout);
    i++;				// Increment counter
    if (i >= MAX_IDENT_RETRY) {		// Too many errors
      fprintf(stderr, "\nNo response from the cable!\n");
      return 1;
     }
   }
  fprintf(stdout, " [ Confirmed ]\n");

  // Read file image

  fp = fopen(argv[ 1 ], "rb");
  if (fp == NULL) {
    fprintf(stderr, "File image open failure!\n");
    return 1;
   }
  len = fread(buf, 1, sizeof(buf), fp);
  if (len == 0) {
    fprintf(stderr, "File image read failure!\n");
    fclose(fp);
    return 1;
   }
  fclose(fp);
  fprintf(stdout, " Source file = %s\n", argv[ 1 ]);
  fprintf(stdout, "Program size = %d bytes\n", len);

  // Transfer file image to GBA

  fprintf(stderr, "Transfering \"%s\" to GBA.\n", argv[ 1 ]);
  for (i = 0; i < 16; i++)
    send_long(OPT_HS_5TH);
  send_long(0);
  send_long((unsigned int) len);

  for (i = 0; i < (len + 3)/4; i++) {
    send_long(*wptr++);
    if (i % 32 == 0) {
      fprintf(stdout, ".");
      fflush(stdout);
     }
   }
  fprintf(stdout, "\n");

  return 0;
 }

// Search and ident the bootcable

int opt_ident(void) {
  int ctr, i;

  ctr = inb_p(CTR);			// Remember current CTR value

  // Device initialization

  outb_p(CTR_STB, CTR);			// Turn on STB --> strobe
					// Turn off INIT --> reset

  outb_p(OPT_HS_1ST, DTR);		// Send 1st handshaking code
  for (i = 0; i < MAX_HS_RETRY; i++) {	// Confirm acknowledge from cable
    if (((inb_p(STR) >> 3) & 0x0F) == OPT_HS_1ST)
      break;
   }
  if (i == MAX_HS_RETRY) {			// Time out error
    i = 1; goto error;
   }

  outb_p(OPT_HS_2ND, DTR);		// Send 2nd handshaking code
  for (i = 0; i < MAX_HS_RETRY; i++) {	// Confirm acknowledge from cable
    if (((inb_p(STR) >> 3) & 0x0F) == OPT_HS_2ND)
      break;
   }
  if (i == MAX_HS_RETRY) {			// Time out error
    i = 2; goto error;
   }

  outb_p(OPT_HS_3RD, DTR);		// Send 3rd handshaking code
  for (i = 0; i < MAX_HS_RETRY; i++) {	// Confirm acknowledge from cable
    if (((inb_p(STR) >> 3) & 0x0F) == OPT_HS_3RD)
      break;
   }
  if (i == MAX_HS_RETRY) {			// Time out error
    i = 3; goto error;
   }
  return 0;

 error:
  outb_p(ctr, CTR);			// Recover original CTR
  return i;				// Return error code
 }

// Receive 4 bits from bootcable

static
int recv_nibble(void) {
  int val;

  outb_p(CTR_INIT | CTR_STB, CTR);	// Turn on strobe --> strobe!
					// INIT = High --> normal mode
  while (inb_p(STR) & STR_BUSY);	// Wait for non-BUSY
  val = (inb_p(STR) >> 3) & 0x0F;	// Read nibble data
  outb_p(0, CTR);			// Turn off STB --> no strobe
					// Turn off INIT --> reset
  while ((inb_p(STR) & STR_BUSY)==0);	// Wait for BUSY
  outb_p(CTR_STB, CTR);			// Turn on strobe --> strobe!
					// Turn off INIT --> reset
  return (val);				// Return the result
 }

// Receive 32 bits data from bootcable

int recv_long(void) {
  int i, res = 0;

  for (i = 0; i < 8; i++)
    res += recv_nibble() << (4 * i);
  return res;
 }

// Send 32 bits to bootcable

void send_long(unsigned int val) {
  int i;

  for (i = 0; i < 8; i++) {
    outb_p((val >> (28 - (4*i))) & 0x0F, DTR);
    outb_p(0, CTR);			// Turn off STB --> no strobe
					// Turn off INIT --> reset
    while (inb_p(STR) & STR_BUSY);	// Wait for non-BUSY
    outb_p(CTR_STB, CTR);		// Turn on strobe --> strobe!
					// Turn off INIT --> reset
   }
  while ((inb_p(STR) & STR_BUSY)==0);	// Wait for BUSY
 }
